use axum::extract::{FromRef, State};
use axum::{routing, Router};
use axum_template::{engine::Engine, Key, RenderHtml};
use minijinja::{path_loader, Environment};
use std::net::SocketAddr;

use serde::Serialize;
use std::sync::{Arc, Mutex};

use vbplay::AudioThread;

mod soundclips;
mod vbplay;

type AppEngine = Engine<Environment<'static>>;

#[derive(Clone)]
pub struct AppState {
    pub engine: AppEngine,
    pub clips: Vec<soundclips::SoundClip>,
    pub player: Arc<Mutex<AudioThread>>,
}

impl FromRef<AppState> for AppEngine {
    fn from_ref(state: &AppState) -> Self {
        state.engine.clone()
    }
}

#[tokio::main]
async fn main() {
    let mut jinja = Environment::new();
    jinja.set_loader(path_loader("templates"));
    let template_engine = Engine::from(jinja);

    let audio = vbplay::audio_thread();
    let audio: Arc<Mutex<AudioThread>> = Arc::new(Mutex::new(audio));

    let state = AppState {
        engine: template_engine,
        clips: soundclips::scan_directory_for_clips("E:/sounds/soundboard", &["mp3"])
            .expect("No Soundclips found."),
        player: audio,
    };

    // Set the address to run our application on
    let addr = SocketAddr::from(([127, 0, 0, 1], 3000));

    // Build our application with a route
    let app = Router::new()
        .route("/", routing::get(home))
        .route("/play/:hash", routing::get(play_handler))
        .route("/stop", routing::get(stop_handler))
        .with_state(state);

    println!("listening on {}", addr);

    // Run it with hyper
    axum::Server::bind(&addr)
        .serve(app.into_make_service())
        .await
        .unwrap();
}

#[derive(Serialize)]
struct TemplateContext {
    clips: Vec<soundclips::SoundClip>,
}

async fn home(engine: AppEngine, state: State<AppState>) -> impl axum::response::IntoResponse {
    let clips = state.0.clone().clips; // this is not ideal.
    let context = TemplateContext { clips: clips };
    RenderHtml(Key("index.html".to_owned()), engine, context)
}

async fn play_handler(
    axum::extract::Path(hash): axum::extract::Path<String>,
    state: State<AppState>,
) -> String {
    if let Some(clip) = state
        .clips
        .clone()
        .into_iter()
        .find(|s| hash == s.hash().to_string())
    {
        let player = state.player.clone();
        let full_file_name = &clip.full_file_name();
        player
            .lock()
            .unwrap()
            .send(vbplay::Command::Play(full_file_name.into()))
            .unwrap();
    }
    "".to_owned()
}

async fn stop_handler(state: State<AppState>) -> String {
    let player = state.player.clone();
    player.lock().unwrap().send(vbplay::Command::Stop).unwrap();

    "".to_owned()
}
